



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
mypy --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
stubgen --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
dmypy --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
stubtest --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
